#include "plotdata.h"
#include "TH2D.h"
#include "TMath.h"
#include "TFile.h"
#include "TStyle.h"
#include "TAxis.h"
#include "TFile.h"
#include <fstream>
#include <iostream>
#include "omp.h"
#include "QDebug"
#include "QDateTime"

PlotData::PlotData(const Setup* config_)
      : config(config_) 
{ 
    counts = 0;
    probDist = std::vector<std::vector<double>>(config->thetaBins, std::vector<double>(config->phiBins, 0));
    pixelateSphere();

    imageData = std::vector<std::vector<double>>(config->thetaBins, std::vector<double>(config->phiBins, 0));
    
    spectrum = new TH1D("Spectrum", " ; Total deposited energy (keV); Counts",
                        config->ergBins, config->ergMin, config->ergMax);
    spectrum->Reset();
    spectrum->SetFillColor(40);

    // Pair
    for(int i=0;i<config->channelSettings.size() / 2;i++)
    {
        std::string id_str = std::string("Pair-") + std::to_string(i);
        TH1D* newhist = new TH1D(id_str.c_str(), " ; Deposited energy (keV); Counts", config->pairErgBins, config->pairErgMin, config->pairErgMax);
        newhist->Reset();
        newhist->SetFillColor(40);
        pairSpectra.push_back(newhist);
    }

    // LG
    for (int i=0;i<128;i++)
    {
        int asic = i/32;
        int channel = i % 32;
        std::string id_str =  std::string("LG-") + std::to_string(asic) + "-" + std::to_string(channel);
        TH1D* newhist = new TH1D(id_str.c_str(), " ; Charge LG; Counts", config->LGHGErgBins, 0, 16384);
        newhist->Reset();
        newhist->SetFillColor(40);
        LGSpectra.push_back(newhist);
    }

    // HG
    for (int i=0;i<128;i++)
    {
        int asic = i/32;
        int channel = i % 32;
        std::string id_str = std::string("HG-") + std::to_string(asic) + "-" + std::to_string(channel);
        TH1D* newhist = new TH1D(id_str.c_str(), " ; Charge HG; Counts", config->LGHGErgBins, 0, 16384);
        newhist->Reset();
        newhist->SetFillColor(40);
        HGSpectra.push_back(newhist);
    }
}

PlotData::~PlotData()
{
    if (spectrum){
        delete spectrum;
        spectrum = nullptr;
    }
    for (std::size_t i=0;i<pairSpectra.size();i++)
    {
        if (pairSpectra[i])
        {
            delete pairSpectra[i];
            pairSpectra[i] = nullptr;
        }
    }
    for (std::size_t i=0;i<LGSpectra.size();i++)
    {
        if(LGSpectra[i])
        {
            delete LGSpectra[i];
            LGSpectra[i] = nullptr;
        }
    }
    for (std::size_t i=0;i<HGSpectra.size();i++)
    {
        if(HGSpectra[i])
        {
            delete HGSpectra[i];
            HGSpectra[i] = nullptr;
        }
    }
}

void PlotData::unZoomSpectrum()
{
    mMutex.lock();
    spectrum->GetXaxis()->UnZoom();
    spectrum->GetYaxis()->UnZoom();
    mMutex.unlock();
}
void PlotData::rebinImage()
{
    mMutex.lock();
    counts = 0;
    probDist = std::vector<std::vector<double>>(config->thetaBins, std::vector<double>(config->phiBins, 0));
    imageData = std::vector<std::vector<double>>(config->thetaBins, std::vector<double>(config->phiBins, 0));
    pixelateSphere();
    mMutex.unlock();
}

void PlotData::rebinSpectrum()
{
    mMutex.lock();
    spectrum->Reset();
    spectrum->SetBins(config->ergBins, config->ergMin, config->ergMax);
    mMutex.unlock();
//    return;
}

void PlotData::rebinLGSpectra()
{
    mMutex.lock();
    for (std::size_t i=0;i<LGSpectra.size();i++)
    {
        LGSpectra[i]->Reset();
        LGSpectra[i]->SetBins(config->LGHGErgBins, 0, 16384);
    }
    mMutex.unlock();
//    return;
}

void PlotData::rebinHGSpectra()
{
    mMutex.lock();
    for (std::size_t i=0;i<HGSpectra.size();i++)
    {
        HGSpectra[i]->Reset();
        HGSpectra[i]->SetBins(config->LGHGErgBins, 0, 16384);
    }
    mMutex.unlock();
//    return;
}

int PlotData::pixelateSphere() {
    xbs = std::vector<std::vector<Vector3D>>(config->thetaBins, std::vector<Vector3D>(config->phiBins, Vector3D(0, 0, 0)));
    thetaBinCenters = std::vector<double>(config->thetaBins, 0);
    phiBinCenters = std::vector<double>(config->phiBins, 0);
    dphi = (config->phiMax - config->phiMin) / config->phiBins;
    dtheta = (config->thetaMax - config->thetaMin) / config->thetaBins;
    double phi = config->phiMin + dphi / 2;;
    double theta = config->thetaMin + dtheta / 2;
    for (int i = 0; i < config->thetaBins; i++)
    {
        thetaBinCenters[i] = theta;
        phi = config->phiMin + dphi / 2;
        for (int j = 0; j < config->phiBins; j++)
        {
            if (i==0)
                phiBinCenters[j] = phi;
            xbs[i][j].X = config->R * std::cos(theta) * std::cos(phi);
            xbs[i][j].Y = config->R * std::cos(theta) * std::sin(phi);
            xbs[i][j].Z = config->R * std::sin(theta);
            phi += dphi;
        }
        theta += dtheta;
    }
    return 0;
}

void PlotData::drawSpectrum()
{
    spectrum->Draw();
}

void PlotData::drawPairSpectrum(const int& pair)
{
    if (pair >= 0 && pair < pairSpectra.size())
        pairSpectra[pair]->Draw();
}
void PlotData::clearPairSpectra()
{
    mMutex.lock();
    for (std::size_t i=0;i<pairSpectra.size();i++)
    {
        pairSpectra[i]->Reset();
    }
    mMutex.unlock();
}

void PlotData::unZoomPairSpectrum(const int& pair)
{
    mMutex.lock();
    pairSpectra[pair]->GetXaxis()->UnZoom();
    pairSpectra[pair]->GetYaxis()->UnZoom();
    mMutex.unlock();
}
bool PlotData::savePairSpectrum2txt(const std::string&fileName, const int& pair)
{
    if (pair < 0 || pair >= pairSpectra.size())
        return false;
    mMutex.lock();
    bool saved=spectrum2txt(pairSpectra[pair], fileName);
    mMutex.unlock();
    return saved;
}
void PlotData::rebinPairSpectra()
{
    mMutex.lock();
    for (std::size_t i=0;i<pairSpectra.size();i++)
    {
        pairSpectra[i]->Reset();
        pairSpectra[i]->SetBins(config->pairErgBins, config->pairErgMin, config->pairErgMax);
    }
    mMutex.unlock();
}
void PlotData::updatePairSpectra(const std::vector<MyPulse>& pulses)
{
    mMutex.lock();
    for (const MyPulse& p : pulses)
    {
        pairSpectra[p.cellNo]->Fill(p.height);
    }
    mMutex.unlock();
}

void PlotData::saveOnExit()
{
    QDateTime dateTime = QDateTime::currentDateTime();
    QString currentDateTime = dateTime.toString("yyyy-MM-dd-HH-mm-ss");
    mMutex.lock();
    if (config->saveImage)
    {
        if (config->saveImageFormat == "ROOT"){
            QString filename = config->outputDir.absolutePath() + QString("/Image-%1.root").arg(currentDateTime);
            TFile f(filename.toLocal8Bit().constData(),"new");
            TH2D hist("2D image", " ; Azimuth (degree); Elevation (degree)",
                      config->phiBins, config->phiMin * 180 / M_PI, config->phiMax * 180 / M_PI,
                      config->thetaBins, config->thetaMin * 180 / M_PI, config->thetaMax * 180 / M_PI);
            for (int i = 0; i < config->thetaBins; i++)
            {
                for (int j = 0; j < config->phiBins; j++)
                {
                    hist.SetBinContent(j+1, i+1, imageData[i][j]);
                }
            }
            hist.Write();
            f.Close();
        }
        else if (config->saveImageFormat == "TXT"){
            QString filename = config->outputDir.absolutePath() + QString("/Image-%1.txt").arg(currentDateTime);
            saveImage2txt(filename.toStdString());
        }
    }
    if (config->saveSpectrum)
    {
        if (config->saveSpectrumFormat == "ROOT"){
            QString filename = config->outputDir.absolutePath() + QString("/Spectrum-%1.root").arg(currentDateTime);
//            spectrum->SaveAs(filename.toLocal8Bit().constData(), "root");
            TFile f(filename.toLocal8Bit().constData(),"new");
            spectrum->Write();
            f.Close();
        }
        else if (config->saveImageFormat == "TXT"){
            QString filename = config->outputDir.absolutePath() + QString("/Spectrum-%1.txt").arg(currentDateTime);
            spectrum2txt(spectrum, filename.toStdString());
        }
    }
    mMutex.unlock();
}

void PlotData::clearAll()
{
    mMutex.lock();
    counts=0;
    for (auto& sub : imageData) {
        std::fill(sub.begin(), sub.end(), 0);
    }
    spectrum->Reset();
    for (std::size_t i=0;i<pairSpectra.size();i++)
        pairSpectra[i]->Reset();
    for (std::size_t i=0;i<LGSpectra.size();i++)
        LGSpectra[i]->Reset();
    for (std::size_t i=0;i<HGSpectra.size();i++)
        HGSpectra[i]->Reset();
    mMutex.unlock();
}

void PlotData::clearImage()
{
    mMutex.lock();
    counts=0;
    for (auto& sub : imageData) {
        std::fill(sub.begin(), sub.end(), 0);
    }
    mMutex.unlock();
}

void PlotData::clearSpectrum()
{
    mMutex.lock();
    spectrum->Reset();
    mMutex.unlock();
}

void PlotData::clearLGSpectrum(const int& asic, const int& channel)
{
    int id = asic * 32 + channel;
    if(id >= LGSpectra.size())
        return;
    mMutex.lock();
    LGSpectra[id]->Reset();
    mMutex.unlock();
}

void PlotData::clearHGSpectrum(const int& asic, const int& channel)
{
    int id = asic * 32 + channel;
    if (id >= HGSpectra.size())
        return;
    mMutex.lock();
    HGSpectra[id]->Reset();
    mMutex.unlock();
}

void PlotData::clearLGHG()
{
    mMutex.lock();
    for (int i=0;i<LGSpectra.size();i++)
    {
        LGSpectra[i]->Reset();
    }

    for (int i=0;i<HGSpectra.size();i++)
    {
        HGSpectra[i]->Reset();
    }
    mMutex.unlock();
}

bool PlotData::saveImage2txt(const std::string& fileName)
{
//        qDebug() << "Save as text file";
    std::ofstream outf(fileName.c_str(), std::ios::out);
    if (!outf.good())
    {
        return false;
    }

    outf << "     Phi   Theta      Content\n";
    const double conv = 180 / M_PI;
    mMutex.lock();
    for (int i = 0; i < config->thetaBins; i++)
    {
        for (int j = 0; j < config->phiBins; j++)
        {
            outf << std::fixed    << std::setprecision(2)
                 << std::setw(8)  << thetaBinCenters[i] * conv
                    << std::setw(8)  << phiBinCenters[j] * conv
                       << std::fixed    << std::setprecision(8)
                       << std::setw(13) << imageData[i][j] << '\n';
        }
    }
    mMutex.unlock();
    outf.close();
    //    qDebug() << "Text file saved.";
    return true;
}

bool PlotData::saveLGSpectrum2txt(const std::string&fileName, const int& asic, const int& channel)
{
    int id = asic * 32 + channel;
    if(id >= LGSpectra.size())
        return false;
    mMutex.lock();
    bool saved=spectrum2txt(LGSpectra[id], fileName);
    mMutex.unlock();
    return saved;
}
bool PlotData::saveHGSpectrum2txt(const std::string&fileName, const int& asic, const int& channel)
{
    int id = asic * 32 + channel;
    if(id >= HGSpectra.size())
        return false;
    mMutex.lock();
    bool saved=spectrum2txt(HGSpectra[id], fileName);
    mMutex.unlock();
    return saved;
}
void PlotData::drawLGspectrum(const int& asic, const int& channel)
{
    int id = asic * 32 + channel;
    if(id >= LGSpectra.size())
        return;
    LGSpectra[id]->Draw();
}
void PlotData::drawHGspectrum(const int& asic, const int& channel)
{
    int id = asic * 32 + channel;
    if(id >= HGSpectra.size())
        return;
    HGSpectra[id]->Draw();
}
void PlotData::unZoomLG(const int& asic, const int& channel)
{
    int id = asic * 32 + channel;
    if(id >= LGSpectra.size())
        return;
    mMutex.lock();
    LGSpectra[id]->GetXaxis()->UnZoom();
    LGSpectra[id]->GetYaxis()->UnZoom();
    mMutex.unlock();
}
void PlotData::unZoomHG(const int& asic, const int& channel)
{
    int id = asic * 32 + channel;
    if(id >= HGSpectra.size())
        return;
    mMutex.lock();
    HGSpectra[id]->GetXaxis()->UnZoom();
    HGSpectra[id]->GetYaxis()->UnZoom();
    mMutex.unlock();
}

bool PlotData::saveSpectrum2txt(const std::string& fileName)
{
//        qDebug() << "Save as text file";
    mMutex.lock();
    bool saved=spectrum2txt(spectrum, fileName);
    mMutex.unlock();
    return saved;
}

bool PlotData::spectrum2txt(const TH1D* s, const std::string& fileName)
{
    std::ofstream outf(fileName.c_str(), std::ios::out);
    if (!outf.good())
    {
        return false;
    }

    outf << "     BinCenter      Counts\n";
    for (int i = 1; i <= s->GetNbinsX(); i++)
    {
        outf << std::fixed    << std::setprecision(2)
             << std::setw(8)  << s->GetBinCenter(i)
             << std::setw(8)  << s->GetBinContent(i)<< '\n';
    }
    outf.close();
//    qDebug() << "Text file saved.";
    return true;
}

void PlotData::updateLGSpectra(const int asic, const ushort (&chargeLG)[32])
{
    mMutex.lock();
    int id = asic * 32;
    for(int i=0;i<32;i++)
    {
        if(chargeLG[i] > 0)
            LGSpectra[id+i]->Fill(chargeLG[i]);
    }
    mMutex.unlock();
}
void PlotData::updateHGSpectra(const int asic, const ushort (&chargeHG)[32])
{
    mMutex.lock();
    int id = asic * 32;
    for(int i=0;i<32;i++)
    {
        if(chargeHG[i] > 0)
            HGSpectra[id+i]->Fill(chargeHG[i]);
    }
    mMutex.unlock();
}

void PlotData::updateImage(std::vector<Cone>::const_iterator first,
                                 std::vector<Cone>::const_iterator last,
                                 const bool normalized)
{
    if (first==last)
        return;
    else if (normalized) {
        mMutex.lock();
        addConesNormalized(first, last);
        mMutex.unlock();
    }
    else {
        mMutex.lock();
        addCones(first, last);
        mMutex.unlock();
    }
}

int PlotData::addCones(std::vector<Cone>::const_iterator first,
             std::vector<Cone>::const_iterator last)
{
    // project cones onto the spherical surface
    double alpha(0);
    double beta(0);
    double E2(0);
    double sgma2(0);
    double sgmb2(0);
    Vector3D ray;
    for (auto k = first; k < last; k++)
    {
        alpha = k->cosHalfAngle;
        spectrum->Fill(k->E0);
        // ideal case: inital energy is known
        // sgma2 = std::pow(0.511*k->E0/std::pow((k->E0-k->Edpst),2)*config->sgmE, 2);
        // realistic case: initial energy = E_1 + E_2
        E2 = k->E0 - k->Edpst;
        sgma2 = std::pow(k->Edpst/std::pow(k->E0, 2) , 2) + std::pow(1/E2 - E2/std::pow(k->E0, 2) , 2);
        sgma2 *= std::pow(config->sgmE * 511, 2);
        #pragma omp parallel for private(ray, beta, sgmb2) collapse(2)
        for (int i = 0; i < config->thetaBins; i++)
        {
            for (int j = 0; j < config->phiBins; j++)
            {
                ray = k->apex - xbs[i][j];
                beta = getCosAngle(ray, k->axis);
                sgmb2 = std::pow((ray/(ray*ray) + k->axis/(k->axis*k->axis)-
                                (ray+k->axis)/(ray*k->axis))*config->sgmpos * beta, 2);
                sgmb2+= std::pow((ray/(ray*k->axis)-k->axis/(k->axis*k->axis))*config->sgmpos * beta, 2);
                imageData[i][j] += std::exp(-std::pow((std::pow(beta, config->order)-std::pow(alpha, config->order)), 2)/
                                            (2*std::pow(config->order, 2)*
                                             (std::pow(alpha,2*config->order-2)*sgma2 +std::pow(beta, 2*config->order-2)*sgmb2)));
            }
        }
    }
    counts += (last - first);
    return 0;
}

//int PlotData::addConesNormalized(std::vector<Cone>::const_iterator first,
//             std::vector<Cone>::const_iterator last)
//{
//    // project cones onto the spherical surface
//    double alpha(0);
//    double beta(0);
//    double E2(0);
//    double sgma2(0);
//    double sgmb2(0);
//    double summ(0);
//    Vector3D ray;
//    for (auto k = first; k < last; k++)
//    {
//        counts++;
//        if (counts == 3507)
//            qDebug() << "here";
//        alpha = k->cosHalfAngle;
//        spectrum->Fill(k->E0);
//        // ideal case: inital energy is known
//        // sgma2 = std::pow(0.511*k->E0/std::pow((k->E0-k->Edpst),2)*config->sgmE, 2);
//        // realistic case: initial energy = E_1 + E_2
//        E2 = k->E0 - k->Edpst;
//        sgma2 = std::pow(k->Edpst/std::pow(k->E0, 2) , 2) + std::pow(1/E2 - E2/std::pow(k->E0, 2) , 2);
//        sgma2 *= std::pow(config->sgmE * 0.511, 2);
//        summ = 0;
//        #pragma omp parallel for private(ray, beta, sgmb2) shared(probDist) reduction(+:summ)
//        for (int i = 0; i < config->thetaBins; i++)
//        {
//            for (int j = 0; j < config->phiBins; j++)
//            {
//                ray = k->apex - xbs[i][j];
//                beta = getCosAngle(ray, k->axis);
//                sgmb2 = std::pow((ray/(ray*ray) + k->axis/(k->axis*k->axis)-
//                                (ray+k->axis)/(ray*k->axis))*config->sgmpos * beta, 2);
//                sgmb2+= std::pow((ray/(ray*k->axis)-k->axis/(k->axis*k->axis))*config->sgmpos * beta, 2);
//                probDist[i][j] = std::exp(-std::pow((std::pow(beta, config->order)-std::pow(alpha, config->order)), 2)/
//                                (2*std::pow(config->order, 2)*
//                                (std::pow(alpha,2*config->order-2)*sgma2 +std::pow(beta, 2*config->order-2)*sgmb2)));
//                // use proability density
//                summ += probDist[i][j] * dtheta * dphi * std::cos(thetaBinCenters[i]);
//                // uncomment to use integral in the bin
//                // probDist[i][j] = probDist[i][j] * config->dtheta * config->dphi * std::cos(config->thetaBinCenters[i]);
//                // summ += probDist[i][j];
//            }
//        }
//        // #pragma omp parallel for
//        for (int i = 0; i < config->thetaBins; i++)
//        {
//            for (int j = 0; j < config->phiBins; j++)
//            {
//                hist->SetBinContent(j+1, i+1, (hist->GetBinContent(j+1, i+1)*(counts - 1) + probDist[i][j] / summ) / counts);
//                if(std::isnan(hist->GetBinContent(j+1, i+1)) || std::isinf(hist->GetBinContent(j+1,i+1)))
//                    qDebug() << i << j << counts;
//            }
//        }
//    }
//    return 0;
//}

int PlotData::addConesNormalized(std::vector<Cone>::const_iterator first,
             std::vector<Cone>::const_iterator last)
{
    // project cones onto the spherical surface
    double alpha(0);
    double beta(0);
    double E2(0);
    double sgma2(0);
//    double sgmb2(0);
//    double summ(0);
    Vector3D ray;
    double A(0);
    for (auto k = first; k < last; k++)
    {
        counts++;
        alpha = k->cosHalfAngle;
        spectrum->Fill(k->E0);
        // ideal case: inital energy is known
        // sgma2 = std::pow(0.511*k->E0/std::pow((k->E0-k->Edpst),2)*config->sgmE, 2);
        // realistic case: initial energy = E_1 + E_2
        E2 = k->E0 - k->Edpst;
        sgma2 = std::pow(k->Edpst/std::pow(k->E0, 2) , 2) + std::pow(1/E2 - E2/std::pow(k->E0, 2) , 2);
        sgma2 *= std::pow(config->sgmE * 511, 2);
        A = 0.5 * M_1_PI * M_2_SQRTPI * M_SQRT1_2 / std::sqrt(sgma2) / (std::erf((1-alpha)/ (M_SQRT2 * std::sqrt(sgma2))) - std::erf((-1-alpha)/ (M_SQRT2 * std::sqrt(sgma2))));
//        #pragma omp parallel for private(ray, beta) shared(probDist)
        for (int i = 0; i < config->thetaBins; i++)
        {
            for (int j = 0; j < config->phiBins; j++)
            {
                ray = k->apex - xbs[i][j];
                beta = getCosAngle(ray, k->axis);

                probDist[i][j] = A * std::exp(-std::pow(beta-alpha, 2)/
                                (2*sgma2));
            }
        }
        // #pragma omp parallel for
        for (int i = 0; i < config->thetaBins; i++)
        {
            for (int j = 0; j < config->phiBins; j++)
            {
                imageData[i][j] = (imageData[i][j] *(counts-1) + probDist[i][j]) / counts;
//                if(std::isnan(hist->GetBinContent(j+1, i+1)) || std::isinf(hist->GetBinContent(j+1,i+1)))
//                    qDebug() << i << j << counts;
            }
        }
    }
    return 0;
}

